from __future__ import annotations

from django.contrib.sites.managers import CurrentSiteManager as BaseCurrentSiteManager
from django.contrib.sites.models import Site
from django.core.exceptions import ObjectDoesNotExist
from django.db import models, transaction


class SiteModelMixinError(Exception):
    pass


class CurrentSiteManager(BaseCurrentSiteManager):
    use_in_migrations = True

    def get_by_natural_key(self, subject_identifier):
        return self.get(subject_identifier=subject_identifier)


class SiteModelMixin(models.Model):
    site = models.ForeignKey(
        Site, on_delete=models.PROTECT, null=True, editable=False, related_name="+"
    )

    on_site = CurrentSiteManager()

    def save(self, *args, **kwargs):
        if not self.id and not self.site:
            self.site = self.get_site_on_create()
        elif "update_fields" in kwargs and "site" not in kwargs.get("update_fields"):
            pass
        else:
            self.validate_site_against_current()
        super().save(*args, **kwargs)

    def get_site_on_create(self) -> Site:
        """Returns a site model instance.

        See also django-multisite.
        """
        site = None
        if not self.site:
            try:
                with transaction.atomic():
                    site = Site.objects.get_current()
            except ObjectDoesNotExist as e:
                raise SiteModelMixinError(e)
        return site or self.site

    def validate_site_against_current(self) -> None:
        """Validate existing site instance matches current_site."""
        pass

    class Meta:
        abstract = True



from django import forms


class SiteModelFormMixin:

    """Validate the current site against a form question.

    This should be used sparingly. A good place is on the screening and/or consent form.

    Declare the modeladmin class with `SiteModelAdminMixin` to have
    the current site set on the form from the request object.

    Declare a `site` field with widget on the ModeForm:

    site = SiteField()

    You will also need to re-declare the `site` model field as `editable`.
    """

    def clean(self) -> dict:
        cleaned_data = super().clean()
        self.validate_with_current_site(cleaned_data)
        return cleaned_data

    def validate_with_current_site(self, cleaned_data: dict) -> None:
        current_site = getattr(self, "current_site", None)
        if (
            current_site
            and cleaned_data.get("site")
            and current_site.id != cleaned_data.get("site").id
        ):
            raise forms.ValidationError(
                {
                    "site": (
                        "Invalid. Please check you are logged into the correct site "
                        "before continuing"
                    )
                }
            )



from .sites import all_sites, sites


class SiteTestCaseMixin:
    @property
    def default_sites(self):
        return sites

    @property
    def site_names(self):
        return [s.name for s in self.default_sites]

    @property
    def default_all_sites(self):
        return all_sites



from __future__ import annotations

from django.conf import settings


class SiteDomainRequiredError(Exception):
    pass


class SiteCountryRequiredError(Exception):
    pass


class SiteLanguagesError(Exception):
    pass


class SingleSite:
    def __init__(
        self,
        site_id: int,
        name: str,
        *,
        title: str = None,
        country: str = None,
        country_code: str = None,
        languages: dict[str, str] | None = None,
        domain: str = None,
        description: str = None,
        fqdn: str = None,
    ):
        if not domain and not fqdn:
            raise ValueError("Require either domain and/or fqdn. Got both as None.")
        self._domain = domain or f"{name}.{fqdn}"
        if languages:
            if unknown_languages := {
                k: v for k, v in languages.items() if k not in self.languages_from_settings
            }:
                raise SiteLanguagesError(
                    f"Unknown language code(s) associated with site. Language code must be "
                    f"included in settings.LANGUAGES. Expected one of settings.LANGUAGES "
                    f"{self.languages_from_settings}. Got {unknown_languages} for site "
                    f"{site_id}."
                )
        self.languages = languages or self.languages_from_settings
        self.site_id = site_id
        self.name = name
        self.title = title or name.title()
        self.country = country or ""
        self.country_code = country_code or ""
        if not country and "multisite" in settings.INSTALLED_APPS:
            raise SiteCountryRequiredError(
                f"Country required when using `multisite`. Got None for `{name}`."
            )
        self.description = description or title

    def __repr__(self):
        return f"{__class__.__name__}(({self.site_id}, {self.domain}, ...))"

    def __str__(self):
        return str(self.domain)

    @property
    def languages_from_settings(self) -> dict[str, str]:
        try:
            lang_iterator = settings.LANGUAGES.items()
        except AttributeError:
            lang_iterator = settings.LANGUAGES
        return {k: v for k, v in lang_iterator}

    @property
    def domain(self) -> str:
        """Returns the domain, inserts `uat` if this is a
        UAT server instance.
        """
        as_list = self._domain.split(".")
        if getattr(settings, "EDC_SITES_UAT_DOMAIN", None):
            if "uat" not in as_list:
                as_list.insert(1, "uat")  # after the site name
        else:
            try:
                as_list.remove("uat")
            except ValueError:
                pass
        self._domain = ".".join(as_list)
        return self._domain

    @property
    def site(self) -> tuple:
        return (
            self.site_id,
            self.name,
            self.title,
            self.country,
            self.domain,
        )

    def as_dict(self) -> dict:
        return dict(
            site_id=self.site_id,
            name=self.name,
            title=self.title,
            country=self.country,
            domain=self.domain,
        )

    def save(self, force_insert=False, force_update=False):
        raise NotImplementedError(f"{self.__class__.__name__} cannot be saved.")

    def delete(self):
        raise NotImplementedError("RequestSite cannot be deleted.")



from .get_site_id import InvalidSiteError, get_sites_from_model


def get_site_name(value, sites=None):
    """Returns the site_name given the site_id."""
    if not sites:
        sites = get_sites_from_model()

    try:
        site_name = [site for site in sites if site.site_id == value][0].name
    except IndexError:
        site_ids = [site.site_id for site in sites]
        raise InvalidSiteError(f"Invalid site. Got '{value}'. Expected one of " f"{site_ids}.")
    return site_name



import json

from django.apps import apps as django_apps

from .single_site import SingleSite


def get_sites_from_model() -> list[SingleSite]:
    sites = []
    site_model_cls = django_apps.get_model("edc_sites.edcsite")
    for obj in site_model_cls.objects.all():
        sites.append(
            SingleSite(
                obj.id,
                obj.name,
                title=obj.title,
                description=obj.description,
                country=obj.country,
                country_code=obj.country_code,
                domain=obj.domain,
                languages=json.loads(obj.languages) if obj.languages else None,
            )
        )
    return sites



from importlib import import_module
from warnings import warn

from django.conf import settings
from django.core.management.color import color_style

style = color_style()


def get_sites_module():
    default_module_name = "edc_sites.sites"
    sites_module_name = getattr(settings, "EDC_SITES_MODULE_NAME", default_module_name)
    if default_module_name == sites_module_name:
        warn(style.NOTICE(f"Using default sites module. `{default_module_name}`."))
    return import_module(sites_module_name or default_module_name)



from django.contrib.sites.models import Site


def get_countries() -> list[str]:
    """Returns the countries"""
    countries = set(s.siteprofile.country for s in Site.objects.all())
    return sorted(list(countries))



from django import forms
from django.conf import settings
from django.contrib.sites.models import Site
from django.test import TestCase
from django.test.utils import override_settings
from multisite import SiteID
from multisite.models import Alias

from edc_sites import get_all_sites, get_current_country, get_sites_by_country
from edc_sites.models import SiteProfile
from edc_sites.sites import all_sites

from ..add_or_update_django_sites import add_or_update_django_sites
from ..forms import SiteModelFormMixin
from ..get_site_id import InvalidSiteError, get_site_id
from ..single_site import SingleSite, SiteLanguagesError
from .models import TestModelWithSite
from .site_test_case_mixin import SiteTestCaseMixin
from .sites import all_test_sites


class TestForm(SiteModelFormMixin, forms.ModelForm):
    class Meta:
        model = TestModelWithSite
        fields = "__all__"


class TestSites(SiteTestCaseMixin, TestCase):
    @override_settings(SITE_ID=SiteID(default=20))
    def test_20(self):
        add_or_update_django_sites(sites=self.default_sites, verbose=False)
        obj = TestModelWithSite.objects.create()
        self.assertEqual(obj.site.pk, 20)
        self.assertEqual(obj.site.pk, Site.objects.get_current().pk)

    @override_settings(SITE_ID=SiteID(default=30))
    def test_30(self):
        add_or_update_django_sites(sites=self.default_sites, verbose=False)
        obj = TestModelWithSite.objects.create()
        self.assertEqual(obj.site.pk, 30)
        self.assertEqual(obj.site.pk, Site.objects.get_current().pk)

    @override_settings(SITE_ID=SiteID(default=30))
    def test_override_current(self):
        add_or_update_django_sites(sites=self.default_sites, verbose=False)
        site = Site.objects.get(pk=40)
        obj = TestModelWithSite.objects.create(site=site)
        self.assertEqual(obj.site.pk, 40)
        self.assertNotEqual(obj.site.pk, Site.objects.get_current().pk)

    def test_get_site_id_by_name(self):
        add_or_update_django_sites(sites=self.default_sites)
        self.assertEqual(get_site_id("mochudi"), 10)

    def test_get_site_id_by_title(self):
        add_or_update_django_sites(sites=self.default_sites)
        self.assertEqual(get_site_id("Mochudi"), 10)

    def test_get_site_id_invalid(self):
        add_or_update_django_sites(sites=self.default_sites)
        self.assertRaises(InvalidSiteError, get_site_id, "blahblah")

    def test_get_site_id_without_sites(self):
        add_or_update_django_sites(sites=self.default_sites)
        self.assertEqual(get_site_id("mochudi"), 10)

    @override_settings(SITE_ID=SiteID(default=30))
    def test_site_profile(self):
        add_or_update_django_sites(sites=self.default_sites, verbose=False)
        obj = TestModelWithSite.objects.create()
        site_profile = SiteProfile.objects.get(site=obj.site)
        self.assertEqual(obj.site.siteprofile, site_profile)

    def test_updates_sites(self):
        add_or_update_django_sites(sites=self.default_sites)
        for site in self.default_sites:
            self.assertIn(site.site_id, [obj.id for obj in Site.objects.all()])
        self.assertNotIn("example.com", str([str(obj) for obj in Site.objects.all()]))
        self.assertEqual(len(self.default_sites), Site.objects.all().count())

    def test_domain(self):
        add_or_update_django_sites(sites=self.default_sites)
        obj = Site.objects.get(name="molepolole")
        self.assertEqual("molepolole.clinicedc.org", obj.domain)
        obj = Site.objects.get(name="mochudi")
        self.assertEqual("mochudi.bw.clinicedc.org", obj.domain)

    @override_settings(EDC_SITES_UAT_DOMAIN=True)
    def test_uat_domain(self):
        self.assertTrue(settings.EDC_SITES_UAT_DOMAIN)
        add_or_update_django_sites(sites=self.default_sites)
        obj = Site.objects.get(name="molepolole")
        self.assertEqual("molepolole.uat.clinicedc.org", obj.domain)
        obj = Site.objects.get(name="mochudi")
        self.assertEqual("mochudi.uat.bw.clinicedc.org", obj.domain)

    @override_settings(SITE_ID=SiteID(default=10))
    def test_country(self):
        for sites in self.default_all_sites.values():
            add_or_update_django_sites(sites=sites)
        self.assertEqual("mochudi", Site.objects.get_current().name)
        self.assertEqual("botswana", Site.objects.get_current().siteprofile.country)
        self.assertEqual("botswana", get_current_country())
        self.assertEqual(
            self.default_all_sites.get("botswana"), get_sites_by_country("botswana")
        )
        self.assertEqual(self.default_all_sites.get("botswana"), get_sites_by_country())
        self.assertEqual(
            self.default_all_sites.get("botswana"),
            get_sites_by_country(all_sites=self.default_all_sites),
        )

        self.assertEqual(
            self.default_all_sites.get("botswana"),
            get_sites_by_country(country="botswana", all_sites=self.default_all_sites),
        )

    @override_settings(EDC_SITES_MODULE_NAME=None)
    def test_default_sites_module_domain(self):
        self.assertEqual(get_all_sites(), all_sites)
        for sites in get_all_sites().values():
            add_or_update_django_sites(sites=sites, verbose=False)
        site = Site.objects.get(id=1)
        self.assertEqual(Alias.objects.get(site=site).domain, "localhost")

    @override_settings(
        EDC_SITES_MODULE_NAME="edc_sites.tests.sites", EDC_SITES_UAT_DOMAIN=False
    )
    def test_custom_sites_module_domain(self):
        self.assertEqual(get_all_sites(), all_test_sites)
        self.assertEqual(settings.SITE_ID, 10)
        for sites in get_all_sites().values():
            add_or_update_django_sites(sites=sites, verbose=False)
        site = Site.objects.get(id=10)
        self.assertEqual(get_current_country(), "botswana")
        self.assertEqual(Alias.objects.get(site=site).domain, "mochudi.bw.clinicedc.org")

    @override_settings(LANGUAGES={"xx": "XXX"})
    def test_languages_not_found(self):
        self.assertRaises(
            SiteLanguagesError,
            SingleSite,
            10,
            "mochudi",
            title="Mochudi",
            country="botswana",
            country_code="bw",
            languages={"tn": "Tswana"},
            domain="mochudi.bw.xxx",
        ),

    @override_settings(LANGUAGES={"tn": "Setswana"})
    def test_languages_ok(self):
        try:
            SingleSite(
                10,
                "mochudi",
                title="Mochudi",
                country="botswana",
                country_code="bw",
                languages={"tn": "Setswana"},
                domain="mochudi.bw.xxx",
            )
        except SiteLanguagesError:
            self.fail("SiteLanguagesError unexpectedly raised")

    @override_settings(LANGUAGES={})
    def test_languages_ok_default(self):
        try:
            SingleSite(
                10,
                "mochudi",
                title="Mochudi",
                country="botswana",
                country_code="bw",
                domain="mochudi.bw.xxx",
            )
        except SiteLanguagesError:
            self.fail("SiteLanguagesError unexpectedly raised")



from __future__ import annotations

import collections

from django.contrib import admin
from django.core.exceptions import FieldError

from ..get_country import get_current_country
from ..get_language_choices_for_site import get_language_choices_for_site


class SiteModeAdminMixinError(Exception):
    pass


class SiteModelAdminMixin:
    language_db_field_name = "language"

    limit_related_to_current_country: list[str] = None
    limit_related_to_current_site: list[str] = None

    @admin.display(description="Site", ordering="site__id")
    def site_code(self, obj=None):
        return obj.site.id

    def get_queryset(self, request):
        """Limit modeladmin queryset for the current site only"""
        qs = super().get_queryset(request)
        if getattr(request, "site", None):
            try:
                qs = qs.filter(site_id=request.site.id)
            except FieldError:
                raise SiteModeAdminMixinError(
                    f"Model missing field `site`. Model `{self.model}`. Did you mean to use "
                    f"the SiteModelAdminMixin? See `{self}`."
                )
        return qs

    def get_form(self, request, obj=None, change=False, **kwargs):
        """Add current_site attr to form instance"""
        form = super().get_form(request, obj=obj, change=change, **kwargs)
        form.current_site = getattr(request, "site", None)
        return form

    def formfield_for_choice_field(self, db_field, request, **kwargs):
        """Use site id to select languages to show in choices."""
        if db_field.name == self.language_db_field_name:
            try:
                language_choices = get_language_choices_for_site(request.site, other=True)
            except AttributeError as e:
                if "WSGIRequest" not in str(e):
                    raise
            else:
                if language_choices:
                    kwargs["choices"] = language_choices
        return super().formfield_for_choice_field(db_field, request, **kwargs)

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        """Filter a ForeignKey field`s queryset by the current site
        or country.

        Note, a queryset set by the ModelForm class will overwrite
        the field's queryset added here.
        """
        self.raise_on_dups_in_field_lists(
            self.limit_related_to_current_country,
            self.limit_related_to_current_site,
        )
        if db_field.name in (self.limit_related_to_current_country or []):
            self.raise_on_queryset_exists(db_field, kwargs)
            country = get_current_country(request)
            model_cls = getattr(self.model, db_field.name).field.related_model
            kwargs["queryset"] = model_cls.objects.filter(siteprofile__country=country)
        elif db_field.name in (self.limit_related_to_current_site or []) and getattr(
            request, "site", None
        ):
            self.raise_on_queryset_exists(db_field, kwargs)
            model_cls = getattr(self.model, db_field.name).field.related_model
            kwargs["queryset"] = model_cls.objects.filter(id=request.site.id)
        elif db_field.name in (self.limit_related_to_current_site or []):
            self.raise_on_queryset_exists(db_field, kwargs)
            model_cls = getattr(self.model, db_field.name).field.related_model
            kwargs["queryset"] = model_cls.on_site.all()
        return super().formfield_for_foreignkey(db_field, request, **kwargs)

    def formfield_for_manytomany(self, db_field, request, **kwargs):
        """Filter a ManyToMany field`s queryset by the current site.

        Note, a queryset set by the ModelForm class will overwrite
        the field's queryset added here.
        """
        self.raise_on_dups_in_field_lists(
            self.limit_related_to_current_country,
            self.limit_related_to_current_site,
        )
        if db_field.name in (self.limit_related_to_current_site or []):
            self.raise_on_queryset_exists(db_field, kwargs)
            model_cls = getattr(self.model, db_field.name).remote_field.model
            kwargs["queryset"] = model_cls.on_site.all()
        elif db_field.name in (self.limit_related_to_current_country or []):
            country = get_current_country(request)
            model_cls = getattr(self.model, db_field.name).remote_field.model
            kwargs["queryset"] = model_cls.objects.filter(siteprofile__country=country)
        return super().formfield_for_manytomany(db_field, request, **kwargs)

    def raise_on_queryset_exists(self, db_field, kwargs):
        """Raise an exception if the `queryset` key exists in the
        kwargs dict.

        If `queryset` exists, remove the field name from the class attr:
            limit_fk_field_to_...
            limit_m2m_field_to_...
        """
        if "queryset" in kwargs:
            raise SiteModeAdminMixinError(
                f"Key `queryset` unexpectedly exists. Got field `{db_field.name}` "
                f"from {self}."
                f". Did you manually set key `queryset` for field `{db_field.name}`?"
            )

    @staticmethod
    def raise_on_dups_in_field_lists(*field_lists: list[str]):
        orig = []
        for field_list in field_lists:
            orig.extend(field_list or [])
        if dups := [item for item, count in collections.Counter(orig).items() if count > 1]:
            raise SiteModeAdminMixinError(
                f"Related field appears in more than one list. Got {dups}."
            )



from edc_sites.single_site import SingleSite

fqdn = "clinicedc.org"
languages = {"en": "English"}
sites: list[SingleSite] = [
    SingleSite(
        10,
        "mochudi",
        title="Mochudi",
        country="botswana",
        country_code="bw",
        languages=languages,
        domain=f"mochudi.bw.{fqdn}",
    ),
    SingleSite(
        20,
        "molepolole",
        title="Molepolole",
        country="botswana",
        country_code="bw",
        languages=languages,
        fqdn=fqdn,
    ),
    SingleSite(
        30,
        "lobatse",
        title="Lobatse",
        country="botswana",
        country_code="bw",
        languages=languages,
        fqdn=fqdn,
    ),
    SingleSite(
        40,
        "gaborone",
        title="Gaborone",
        country="botswana",
        country_code="bw",
        languages=languages,
        fqdn=fqdn,
    ),
    SingleSite(
        50,
        "karakobis",
        title="Karakobis",
        country="botswana",
        country_code="bw",
        languages=languages,
        fqdn=fqdn,
    ),
    SingleSite(
        60,
        "windhoek",
        title="Windhoek",
        country="namibia",
        country_code="na",
        languages=languages,
        fqdn=fqdn,
    ),
]


more_sites = [
    SingleSite(
        60,
        "windhoek",
        title="Windhoek",
        country="namibia",
        country_code="na",
        languages=languages,
        fqdn=fqdn,
    ),
]

all_sites = {"botswana": sites, "namibia": more_sites}
all_test_sites = {"botswana": sites, "namibia": more_sites}



from __future__ import annotations

from django.contrib.sites.models import Site
from django.db import models

from .model_mixins import CurrentSiteManager, SiteModelMixin  # noqa


class SiteModelError(Exception):
    pass


class SiteProfile(models.Model):
    id = models.BigAutoField(primary_key=True)

    country = models.CharField(max_length=250, null=True)

    country_code = models.CharField(max_length=15, null=True)

    languages = models.TextField(null=True)

    title = models.CharField(max_length=250, null=True)

    description = models.TextField(null=True)

    site = models.OneToOneField(Site, on_delete=models.PROTECT)

    def __str__(self):
        return f"{self.site.id}: {self.title}"


class EdcSite(Site):
    @property
    def title(self) -> str:
        return SiteProfile.objects.get(site=self).title

    @property
    def description(self) -> str:
        return SiteProfile.objects.get(site=self).description

    @property
    def country(self) -> str:
        return SiteProfile.objects.get(site=self).country

    @property
    def country_code(self) -> str:
        return SiteProfile.objects.get(site=self).country_code

    @property
    def languages(self) -> str | None:
        return SiteProfile.objects.get(site=self).languages

    class Meta:
        proxy = True



from __future__ import annotations

from django.core.exceptions import ObjectDoesNotExist

from .models import SiteProfile


class SiteViewMixin:
    def get_context_data(self, **kwargs) -> dict:
        context = super().get_context_data(**kwargs)
        try:
            site_profile = SiteProfile.objects.get(site__id=self.request.site.id)
        except ObjectDoesNotExist:
            context.update(site_profile=None)
        else:
            context.update(site_title=site_profile.title)
            context.update(site_profile=site_profile)
        return context



from django.db import models

from ..models import CurrentSiteManager, SiteModelMixin


class TestModelWithSite(SiteModelMixin, models.Model):
    f1 = models.CharField(max_length=10, default="1")

    objects = models.Manager()

    on_site = CurrentSiteManager()



from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist


class EdcSitesCountryError(Exception):
    pass


def get_current_country(request=None):
    """Returns the country, defaults to that of the default site."""
    country = None
    if site := getattr(request, "site", None):
        country = site.siteprofile.country
    else:
        site_model_cls = django_apps.get_model("sites.site")
        try:
            country = site_model_cls.objects.get_current().siteprofile.country
        except ObjectDoesNotExist:
            pass
    return country



"""Default sites module.

Define ``sites.py`` in your own module and set EDC_SITES_MODULE_NAME
to the name of that module.

"""
from django.conf import settings

from edc_sites.single_site import SingleSite

default_country: str = getattr(settings, "EDC_SITES_DEFAULT_COUNTRY", "botswana")
default_country_code: str = getattr(settings, "EDC_SITES_DEFAULT_COUNTRY_CODE", "bw")
default_domain: str = getattr(settings, "EDC_SITES_DEFAULT_DOMAIN", "localhost")

all_sites: dict[str, list[SingleSite]] = {
    default_country
    or "site": [
        SingleSite(
            1,
            settings.APP_NAME,
            country=default_country,
            country_code=default_country_code,
            domain=default_domain,
        ),
    ],
}



from .get_sites_from_model import get_sites_from_model


class InvalidSiteError(Exception):
    pass


def get_site_id(value, sites=None):
    """Returns the site_id given the site_name"""
    if not sites:
        sites = get_sites_from_model()

    try:
        site_id = [site for site in sites if site.name == value][0].site_id
    except IndexError:
        try:
            site_id = [site for site in sites if site.title == value][0].site_id
        except IndexError:
            site_ids = [site.site_id for site in sites]
            site_names = [site.name for site in sites]
            raise InvalidSiteError(
                f"Invalid site. Got '{value}'. Expected one of " f"{site_ids} or {site_names}."
            )
    return site_id



from .add_or_update_django_sites import add_or_update_django_sites
from .get_all_sites import get_all_sites
from .get_country import get_current_country
from .get_site_by_attr import get_site_by_attr
from .get_site_id import InvalidSiteError, get_site_id
from .get_site_name import get_site_name
from .get_sites_by_country import get_sites_by_country
from .get_sites_module import get_sites_module
from .valid_site_for_subject_or_raise import (
    InvalidSiteForSubjectError,
    valid_site_for_subject_or_raise,
)



from __future__ import annotations

import json

from django.apps import apps as django_apps
from edc_constants.constants import OTHER


def get_language_choices_for_site(site, other=None) -> tuple | None:
    """Returns a choices tuple of languages from the site object to
    be used on the `languages` modelform field.

    See also: SingleSite and SiteModelAdminMixin.
    """
    site_profile_model_cls = django_apps.get_model("edc_sites.siteprofile")
    obj = site_profile_model_cls.objects.get(site=site)
    if obj.languages:
        languages = json.loads(obj.languages)
        if other:
            languages.update({OTHER: "Other"})
        return tuple((k, v) for k, v in languages.items())
    return None



from __future__ import annotations

import json
import sys
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist

from .get_sites_module import get_sites_module
from .single_site import SiteDomainRequiredError

if TYPE_CHECKING:
    from .models import SiteProfile


class InvalidSiteError(Exception):
    pass


def add_or_update_django_sites(apps=None, sites=None, verbose=None):
    """Removes default site and adds/updates given `sites`, etc.

    Title is stored in SiteProfile.

    kwargs:
        * sites: format
            sites = (
                (<site_id>, <site_name>, <title>),
                ...)
    """
    all_sites = {}
    if verbose:
        sys.stdout.write("  * updating sites.\n")
    apps = apps or django_apps
    site_model_cls = apps.get_model("sites", "Site")
    site_model_cls.objects.filter(name="example.com").delete()
    if not sites:
        all_sites = get_sites_module().all_sites
    elif isinstance(sites, (list, tuple)):
        all_sites = {"default": sites}
    for sites in all_sites.values():
        for single_site in sites:
            if get_sites_module() and single_site.name == "edc_sites.sites":
                continue
            if verbose:
                sys.stdout.write(f"  * {single_site.name}.\n")
            site_obj = get_or_create_site_obj(single_site, apps)
            get_or_create_site_profile_obj(single_site, site_obj, apps)
    return all_sites


def get_or_create_site_obj(single_site, apps):
    if "multisite" in settings.INSTALLED_APPS and not single_site.domain:
        raise SiteDomainRequiredError(
            f"Domain required when using `multisite`. Got None for `{single_site.name}`"
        )
    site_model_cls = apps.get_model("sites", "Site")
    try:
        site_obj = site_model_cls.objects.get(pk=single_site.site_id)
    except ObjectDoesNotExist:
        site_obj = site_model_cls.objects.create(
            pk=single_site.site_id, name=single_site.name, domain=single_site.domain
        )
    else:
        site_obj.name = single_site.name
        site_obj.domain = single_site.domain
        site_obj.save()
    return site_obj


def get_or_create_site_profile_obj(single_site, site_obj, apps) -> SiteProfile | None:
    site_profile_model_cls = apps.get_model("edc_sites", "SiteProfile")
    opts = dict(
        title=single_site.title,
        country=single_site.country,
        country_code=single_site.country_code,
        description=single_site.description or single_site.title,
        languages=json.dumps(single_site.languages) if single_site.languages else None,
    )
    try:
        site_profile = site_profile_model_cls.objects.get(site=site_obj)
    except ObjectDoesNotExist:
        site_profile = site_profile_model_cls.objects.create(site=site_obj, **opts)
    else:
        for k, v in opts.items():
            setattr(site_profile, k, v)
        site_profile.save()
    return site_profile



from __future__ import annotations

from typing import TYPE_CHECKING

from django.core.exceptions import ObjectDoesNotExist
from edc_registration import get_registered_subject_model_cls

from edc_sites.get_site_model_cls import get_site_model_cls

if TYPE_CHECKING:
    from django.contrib.sites.models import Site


class InvalidSiteForSubjectError(Exception):
    pass


def valid_site_for_subject_or_raise(subject_identifier: str) -> Site:
    """Raises an InvalidSiteError exception if the subject_identifier is not
    from the current site.

    * Confirms by querying RegisteredSubject.
    * If subject_identifier is invalid will raise ObjectDoesNotExist
    """
    current_site = get_site_model_cls().objects.get_current()
    try:
        get_registered_subject_model_cls().objects.get(
            site=current_site, subject_identifier=subject_identifier
        )
    except ObjectDoesNotExist:
        try:
            obj = get_registered_subject_model_cls().objects.get(
                subject_identifier=subject_identifier
            )
        except ObjectDoesNotExist as e:
            raise InvalidSiteForSubjectError(
                "Unable to validate site for subject. subject_identifier="
                f"`{subject_identifier}`.  Got `{e}`"
            )
        else:
            raise InvalidSiteForSubjectError(
                f"Invalid site for subject. {subject_identifier}. Expected `{obj.site.name}`. "
                f"Got `{current_site.name}`"
            )
    return current_site



from __future__ import annotations

from typing import TYPE_CHECKING

from django_extensions.management.color import color_style

from .get_all_sites import get_all_sites
from .get_country import get_current_country
from .get_sites_module import get_sites_module

if TYPE_CHECKING:
    from .single_site import SingleSite

style = color_style()


def get_sites_by_country(
    country: str = None, all_sites: dict[str, list[SingleSite]] | None = None
) -> list[SingleSite]:
    """Returns a sites tuple for the country."""
    if not all_sites or get_all_sites():
        all_sites = get_sites_module().all_sites
    return all_sites.get(country or get_current_country())



from edc_sites.get_sites_from_model import get_sites_from_model


class SiteDoesNotExist(Exception):
    pass


def get_site_by_attr(attrname, value, sites=None):
    if not sites:
        sites = get_sites_from_model()
    try:
        site = [site for site in sites if getattr(site, attrname) == value][0]
    except IndexError:
        raise SiteDoesNotExist(f"No site exists with `{attrname}`=={value}.")
    return site
